"""Define a SmartCocoon Fan class."""

import logging
from datetime import datetime
from typing import Any, Optional

from pysmartcocoon.api import SmartCocoonAPI
from pysmartcocoon.const import FanMode
from pysmartcocoon.fan_helpers import derive_mode_from_speed, resolve_speed

_LOGGER: logging.Logger = logging.getLogger(__name__)


# pylint: disable=too-many-instance-attributes, too-many-public-methods
class Fan:
    """Define the fan."""

    def __init__(
        self,
        fan_id: str,
        api: SmartCocoonAPI,
    ) -> None:
        """Initialize."""

        # Fan attributes from SmartCocoon
        self._fan_id: str = fan_id
        self._identifier: Optional[int] = None
        self._fan_on: Optional[bool] = None
        self._firmware_version: Optional[str] = None
        self._is_room_estimating: Optional[bool] = None
        self._connected: Optional[bool] = None
        self._last_connection: Optional[datetime] = None
        self._mode: Optional[str] = None
        self._power: Optional[int] = None
        self._predicted_room_temperature: Optional[float] = None
        self._room_id: Optional[int] = None
        self._thermostat_vendor: Optional[int] = None
        self._mqtt_username: Optional[str] = None
        self._mqtt_password: Optional[str] = None

        # Extra attributes not provided by API
        self._room_name: Optional[str] = None

        self._api = api

    @property
    def identifier(self) -> Optional[int]:  # pylint: disable=invalid-name
        """Return Fan id.

        This is a numerical ID generated by SmartCocoon
        This value can change if the fan is re-added to
        an cloud account
        """
        return self._identifier

    @property
    def fan_id(self) -> str:
        """Return Fan fan_id.

        This is the physical ID printed on the fan
        """
        return self._fan_id

    @property
    def fan_on(self) -> Optional[bool]:
        """Return a bool indicating if the fan is on or off
        On = True
        Off = False
        """
        return self._fan_on

    @property
    def firmware_version(self) -> Optional[str]:
        """Return the firmware version of the fan"""
        return self._firmware_version

    @property
    def is_room_estimating(self) -> Optional[bool]:
        """Return value to indicate if the room is estimating temp
        Estimating = True
        Not Estimating = False
        """
        return self._is_room_estimating

    @property
    def connected(self) -> Optional[bool]:
        """Return a bool indicating if the fan is connected
        Connected = True
        Not Connected = False
        """
        return self._connected

    @property
    def last_connection(self) -> Optional[datetime]:
        """Return a bool indicating if the fan is connected
        Connected = True
        Not Connected = False
        """
        return self._last_connection

    @property
    def mode(self) -> Optional[str]:
        """Return the current mode setting of the fan"""
        return self._mode

    @property
    def mode_enum(self) -> FanMode:
        """Return the current mode setting of the fan"""
        return FanMode(self._mode) if self._mode is not None else FanMode.OFF

    def get_extra_state_attributes(self) -> dict[str, Any]:
        """Return extra state attributes for Home Assistant integration."""
        attributes = {
            "mode": self._mode,
            "connected": self._connected,
            "last_connection": (
                self._last_connection.isoformat()
                if self._last_connection
                else None
            ),
            "firmware_version": self._firmware_version,
            "power": self._power,
            "room_name": self._room_name,
            "room_id": self._room_id,
        }

        # Add time since last connection
        if self._last_connection:
            time_since = (
                datetime.now(self._last_connection.tzinfo)
                - self._last_connection
            )
            attributes["time_since_connection"] = str(time_since).split(
                ".", maxsplit=1
            )[
                0
            ]  # Remove microseconds
            attributes["connection_status"] = (
                "Connected" if self._connected else "Disconnected"
            )
        else:
            attributes["time_since_connection"] = "Unknown"
            attributes["connection_status"] = "Unknown"

        return attributes

    @property
    def power(self) -> int:
        """Return the current power setting of the fan
        % value is multiplied by 100
        100% = power level of 10000
        """
        return self._power or 0

    @property
    def speed_pct(self) -> int:
        """Return the current power setting of the fan
        by %
        """
        return int(self._power / 100) if self._power is not None else 0

    @property
    def predicted_room_temperature(self) -> Optional[float]:
        """Return the predicted room temperature of the fan"""
        return self._predicted_room_temperature

    @property
    def room_name(self) -> Optional[str]:
        """Return the room name of the fan"""
        return self._room_name

    @property
    def room_id(self) -> Optional[int]:
        """Return the room id of the fan"""
        return self._room_id

    @property
    def thermostat_vendor(self) -> Optional[int]:
        """Return the thermostat vendor of the fan"""
        return self._thermostat_vendor

    @property
    def mqtt_username(self) -> Optional[str]:
        """Return the MQTT username of the fan"""
        return self._mqtt_username

    @property
    def mqtt_password(self) -> Optional[str]:
        """Return the MQTT password of the fan"""
        return self._mqtt_password

    def set_speed_pct(self, fan_speed_pct: int) -> bool:
        """Update the power of the fan"""

        if fan_speed_pct > 100:
            _LOGGER.debug(
                (
                    "Fan ID: %s - Fan speed of %s%% is invalid, must be "
                    "between 0%% and 100%%"
                ),
                self.fan_id,
                str(fan_speed_pct),
            )

        _LOGGER.debug(
            "Fan ID: %s - Updating fan speed to %s%%",
            self.fan_id,
            str(fan_speed_pct),
        )

        self._power = fan_speed_pct * 100
        return True

    def set_room_name(self, room_name: str) -> bool:
        """Update the room_name"""

        _LOGGER.debug(
            "Fan ID: %s - Updating room_name to '%s'", self.fan_id, room_name
        )

        self._room_name = room_name
        return True

    # pylint: disable=too-many-branches
    async def async_set_fan_modes(
        self,
        fan_mode: Optional[FanMode] = None,
        fan_speed_pct: Optional[int] = None,
    ) -> bool:
        """Set the fan mode and speed."""

        _LOGGER.debug(
            (
                "Fan ID: %s - In async_set_fan_mode with fan_mode: %s and"
                " fan_speed: %s%%"
            ),
            self.fan_id,
            fan_mode,
            str(fan_speed_pct),
        )

        if fan_mode is None and fan_speed_pct is None:
            _LOGGER.debug(
                "async_set_fan_modes must provide a value for fan_mode and/or"
                " fan_speed"
            )
            return False

        if fan_mode is None:
            fan_mode = derive_mode_from_speed(self.mode_enum, fan_speed_pct)

        # Update fan mode if changed
        if self.mode_enum != fan_mode:
            self._mode = fan_mode.value

        fan_speed_pct = resolve_speed(self.speed_pct, fan_mode, fan_speed_pct)
        if fan_speed_pct is None:
            return False

        # Update power if changed
        if self.speed_pct != fan_speed_pct:
            self.set_speed_pct(fan_speed_pct)

        # Attempt to update the fan via API
        success = await self._async_set_fan(fan_mode)
        return success

    # helpers moved to fan_helpers.py

    async def _async_set_fan(self, fan_mode: Optional[FanMode] = None) -> bool:
        """Call the API to update the fan mode and speed."""

        # Check if we have the required data to make the API call
        if self._identifier is None:
            _LOGGER.warning(
                "Fan ID: %s - Cannot update fan: identifier is None",
                self.fan_id,
            )
            return False

        if self.mode is None:
            _LOGGER.warning(
                "Fan ID: %s - Cannot update fan: mode is None", self.fan_id
            )
            return False

        # Make the API call
        await self._api.async_update_fan(
            fan_identifier=self._identifier,
            mode=self.mode,
            power=self.power,
        )

        _LOGGER.debug(
            "Fan ID: %s - Fan Mode was set to %s, speed to %s",
            self.fan_id,
            self.mode,
            self.speed_pct,
        )

        await self._async_update_fan()

        if fan_mode == FanMode.ON and not self.fan_on:
            _LOGGER.debug(
                "Fan ID: %s - Changing fan_on to 'True'", self.fan_id
            )
            self._fan_on = True
        elif fan_mode == FanMode.OFF:
            _LOGGER.debug(
                "Fan ID: %s - Changing fan_on to 'False'", self.fan_id
            )
            self._fan_on = False

        return True

    async def async_update_api_data(
        self,
        data: dict[str, Any],
    ) -> bool:
        """Selectively update the fan attributes with API data"""

        _LOGGER.debug("Fan ID: %s - In async_update_api_data", data["fan_id"])

        self._identifier = data["id"]
        self._fan_id = (
            data["fan_id"] if data.get("fan_id") is not None else self._fan_id
        )

        # Fan attributes from SmartCocoon
        self._mode = data["mode"]

        # fan_on does not always reflect the current mode, mode is more
        # accurate if set to always_on or always_off

        if self._mode == FanMode.ON.value:
            self._fan_on = True
        elif self._mode == FanMode.OFF.value:
            self._fan_on = False
        else:
            self._fan_on = data["fan_on"]

        self._firmware_version = data["firmware_version"]
        self._is_room_estimating = data["is_room_estimating"]
        # Log raw connected value from API for debugging
        raw_connected = data.get("connected")
        _LOGGER.debug(
            "Fan ID: %s - Raw API 'connected' value: %s (type: %s)",
            self.fan_id,
            raw_connected,
            type(raw_connected).__name__,
        )

        # Parse last_connection to datetime when provided as string
        last_conn = data["last_connection"]
        if isinstance(last_conn, str):
            try:
                iso_str = last_conn.replace("Z", "+00:00")
                self._last_connection = datetime.fromisoformat(iso_str)
            except Exception:  # pylint: disable=broad-except
                _LOGGER.debug(
                    "Fan ID: %s - Unable to parse last_connection: %s",
                    self.fan_id,
                    last_conn,
                )
                self._last_connection = None
        else:
            self._last_connection = last_conn

        # Correct stale connection status from API
        # If API says connected=True but last_connection is > 15 minutes ago,
        # the fan is actually disconnected (API has stale data)
        api_connected = data["connected"]
        if api_connected and self._last_connection:
            time_since_connection = (
                datetime.now(self._last_connection.tzinfo)
                - self._last_connection
            )
            minutes_since_connection = (
                time_since_connection.total_seconds() / 60
            )

            # If fan hasn't been seen in 15+ minutes, it's disconnected
            if minutes_since_connection > 15:
                _LOGGER.warning(
                    "Fan ID: %s - API reports connected=True but "
                    "last_connection was %.1f minutes ago. "
                    "Overriding to connected=False",
                    self.fan_id,
                    minutes_since_connection,
                )
                self._connected = False
            else:
                self._connected = api_connected
        elif api_connected and self._last_connection is None:
            # API says connected but no last_connection timestamp - trust API
            _LOGGER.debug(
                "Fan ID: %s - API reports connected=True but no "
                "last_connection timestamp available",
                self.fan_id,
            )
            self._connected = api_connected
        else:
            # API says connected=False, trust it
            self._connected = api_connected
        self._power = data["power"]
        self._predicted_room_temperature = data["predicted_room_temperature"]
        self._room_id = data["room_id"]
        self._thermostat_vendor = data["thermostat_vendor"]
        self._mqtt_username = data["mqtt_username"]
        self._mqtt_password = data["mqtt_password"]

        return True

    async def _async_update_fan(self) -> bool:
        _LOGGER.debug(
            "Fan ID: %s - Updating fan attributes from cloud", self.fan_id
        )

        if self._identifier is None:
            _LOGGER.warning(
                "Fan ID: %s - Cannot update fan: identifier is None",
                self.fan_id,
            )
            return False

        response = await self._api.async_get_fan(self._identifier)

        if response is None:
            _LOGGER.warning(
                "Fan ID: %s - Failed to get fan data from API", self.fan_id
            )
            return False

        await self.async_update_api_data(response)
        return True



"""Tests SmartCocoon api module.

Note: Integration tests in this file are marked with @pytest.mark.integration
and will NOT run in GitHub Actions CI. They require real API credentials
and are only intended for local development and testing.
"""

import asyncio
import logging
import sys
from os import environ
from os.path import dirname, join

import aiohttp
import pytest
from dotenv import load_dotenv

from pysmartcocoon.const import FanMode
from pysmartcocoon.manager import SmartCocoonManager

# Removed unused cast import


# Load the account information
dotenv_path = join(dirname(__file__), ".env")
load_dotenv(dotenv_path)

# Environment variables will be checked inside test functions

logging.basicConfig(level=logging.DEBUG)
_LOGGER: logging.Logger = logging.getLogger(__name__)


if environ.get("RUN_INTEGRATION") != "1":
    pytestmark = pytest.mark.skip(
        reason="Set RUN_INTEGRATION=1 to run integration tests"
    )


async def _fan_control(user: str, pwd: str, fid: str) -> None:
    """Test authentication then control fan."""

    async with aiohttp.ClientSession() as session:
        # Init manager
        manager = SmartCocoonManager(session=session)

        if not await manager.async_start_services(username=user, password=pwd):
            _LOGGER.debug("Authentication failed")
            return

        # Test fan controls
        wait_for = 5

        await manager.async_fan_turn_on(fid)
        await asyncio.sleep(wait_for)
        await manager.async_set_fan_speed(fid, 50)
        await asyncio.sleep(wait_for)
        await manager.async_set_fan_speed(fid, 100)
        await asyncio.sleep(wait_for)
        await manager.async_set_fan_auto(fid)
        await asyncio.sleep(wait_for)
        await manager.async_set_fan_eco(fid)
        await asyncio.sleep(wait_for)
        await manager.async_fan_turn_off(fid)
        await asyncio.sleep(wait_for)


@pytest.mark.integration
def test_integration_fan_control() -> None:
    """Integration smoke: fan control flow."""
    # Check for required environment variables
    test_username = environ.get("USERNAME")
    test_password = environ.get("PASSWORD")
    test_fan_id = environ.get("FAN_ID")

    if not test_username or not test_password or not test_fan_id:
        pytest.skip(
            "Missing required environment variables: "
            "USERNAME, PASSWORD, FAN_ID"
        )

    asyncio.run(_fan_control(test_username, test_password, test_fan_id))


async def _fan_modes(user: str, pwd: str, fid: str) -> None:
    """Smoke test: set fan modes and verify reported state."""

    async with aiohttp.ClientSession() as session:
        manager = SmartCocoonManager(session=session)

        if not await manager.async_start_services(username=user, password=pwd):
            _LOGGER.debug("Authentication failed")
            return

        wait_for = int(environ.get("INTEGRATION_WAIT", 3))

        await manager.async_fan_turn_on(fid)
        await asyncio.sleep(wait_for)
        await manager.async_update_fans()
        assert manager.fans[fid].mode_enum == FanMode.ON

        await manager.async_set_fan_speed(fid, 33)
        await asyncio.sleep(wait_for)
        await manager.async_update_fans()
        # Mode should remain ON when changing speed
        assert manager.fans[fid].mode_enum == FanMode.ON

        await manager.async_set_fan_auto(fid)
        await asyncio.sleep(wait_for)
        await manager.async_update_fans()
        assert manager.fans[fid].mode_enum == FanMode.AUTO

        await manager.async_set_fan_eco(fid)
        await asyncio.sleep(wait_for)
        await manager.async_update_fans()
        assert manager.fans[fid].mode_enum == FanMode.ECO

        await manager.async_fan_turn_off(fid)
        await asyncio.sleep(wait_for)
        await manager.async_update_fans()
        assert manager.fans[fid].mode_enum == FanMode.OFF


@pytest.mark.integration
def test_integration_fan_modes() -> None:
    """Integration smoke: verify fan mode transitions."""
    # Check for required environment variables
    test_username = environ.get("USERNAME")
    test_password = environ.get("PASSWORD")
    test_fan_id = environ.get("FAN_ID")

    if not test_username or not test_password or not test_fan_id:
        pytest.skip(
            "Missing required environment variables: "
            "USERNAME, PASSWORD, FAN_ID"
        )

    asyncio.run(_fan_modes(test_username, test_password, test_fan_id))


async def _test_debug_logging(user: str, pwd: str) -> None:
    """Test debug logging functionality."""
    _LOGGER.info("=== Testing Debug Logging Functionality ===")

    async with aiohttp.ClientSession() as session:
        manager = SmartCocoonManager(session=session)

        _LOGGER.info(
            "Manager created - debug logging should show detailed API calls"
        )
        _LOGGER.info("Attempting authentication (watch for debug output)...")

        # This should trigger debug logging for the authentication API call
        auth_result = await manager.async_start_services(
            username=user, password=pwd
        )

        if auth_result:
            _LOGGER.info(
                "Authentication successful - debug logging should have shown:"
            )
            _LOGGER.info("- API REQUEST details (method, URL, headers, body)")
            _LOGGER.info("- API RESPONSE details (status, headers, body)")
            _LOGGER.info("- AUTHENTICATION SUCCESS details (user info, token)")

            # Test a simple API call to show more debug output
            _LOGGER.info(
                "Testing fan data retrieval (watch for more debug output)..."
            )
            await manager.async_update_fans()
            _LOGGER.info(
                "Fan data retrieved - debug logging should have shown details"
            )
        else:
            _LOGGER.warning(
                "Authentication failed - check debug output for error details"
            )


@pytest.mark.integration
def test_integration_debug_logging() -> None:
    """Integration test: verify debug logging functionality."""
    # Check for required environment variables
    test_username = environ.get("USERNAME")
    test_password = environ.get("PASSWORD")
    test_fan_id = environ.get("FAN_ID")

    if not test_username or not test_password or not test_fan_id:
        pytest.skip(
            "Missing required environment variables: "
            "USERNAME, PASSWORD, FAN_ID"
        )

    asyncio.run(_test_debug_logging(test_username, test_password))


if __name__ == "__main__":
    # Allow manual execution outside pytest
    username = environ.get("USERNAME")
    password = environ.get("PASSWORD")
    fan_id = environ.get("FAN_ID")

    if not username or not password or not fan_id:
        print(
            "Missing required environment variables: "
            "USERNAME, PASSWORD, FAN_ID"
        )
        sys.exit(1)

    asyncio.run(_test_debug_logging(username, password))
    asyncio.run(_fan_control(username, password, fan_id))
    asyncio.run(_fan_modes(username, password, fan_id))



"""Define a manager to interact with SmartCocoon"""

import asyncio
import json
import logging
import random
from datetime import datetime, timedelta
from typing import Any, Optional, cast

import async_timeout
from aiohttp import ClientResponseError, ClientSession, ClientTimeout
from aiohttp.client_exceptions import ClientConnectionError

from pysmartcocoon.const import (
    API_AUTH_URL,
    API_FANS_URL,
    API_HEADERS,
    DEFAULT_TIMEOUT,
)
from pysmartcocoon.errors import RequestError, UnauthorizedError

_LOGGER: logging.Logger = logging.getLogger(__name__)


# pylint: disable=too-many-instance-attributes
class SmartCocoonAPI:
    """This class will communicate with the SmartCocoon cloud API"""

    def __init__(
        self,
        session: Optional[ClientSession] = None,
        request_timeout: int = DEFAULT_TIMEOUT,
    ) -> None:
        self._session = session
        self._request_timeout = request_timeout
        self._authenticated = False
        self._api_client: Optional[str] = None
        self._bearer_token: Optional[str] = None
        self._bearer_token_expiration: Optional[datetime] = None
        # Make a private copy of default headers to avoid global mutation
        self._headers_auth = API_HEADERS.copy()
        self._user_id: Optional[int] = None

    async def __aenter__(self) -> "SmartCocoonAPI":
        return self

    async def __aexit__(self, *_: Any) -> None:
        await self.close()

    async def close(self) -> None:
        """Close internally-owned session if present."""
        if self._session and not self._session.closed:
            await self._session.close()

    async def async_authenticate(self, username: str, password: str) -> bool:
        """Function to authenticate user with API"""
        self._authenticated = False

        # Authenticate with user and pass
        request_body: dict[str, Any] = {}
        request_body.setdefault("json", {})
        request_body["json"]["email"] = username
        request_body["json"]["password"] = password

        await self.async_request("POST", API_AUTH_URL, **request_body)

        return self._authenticated

    def _compute_retry_delay(
        self, attempt: int, retry_after: str | None
    ) -> float:
        """Compute delay before retrying, honoring Retry-After when present.

        Uses exponential backoff with jitter when header is absent or invalid.
        """
        if retry_after and retry_after.isdigit():
            return float(int(retry_after))
        base = 2 ** (attempt - 1)
        return base + random.uniform(0, 0.5 * base)

    # pylint: disable=too-many-branches,too-many-statements
    async def async_request(
        self, method: str, url: str, **kwargs: Any
    ) -> dict | None:
        """Make a request using token authentication.
        Args:
            method: Method for the HTTP request (example "GET" or "POST").
            path: path of the REST API endpoint.
        Returns:
            the Response object corresponding to the result of the API request.
        """
        # pylint: disable=broad-except
        use_running_session = self._session and not self._session.closed

        if use_running_session:
            session = self._session
        else:
            timeout_value = ClientTimeout(total=self._request_timeout)
            session = ClientSession(timeout=timeout_value)

        assert session

        data = None

        _LOGGER.debug(
            "Calling SmartCocoon API - method: %s, url: %s", method, url
        )

        # Enhanced debug logging for HA integration
        if _LOGGER.isEnabledFor(logging.DEBUG):
            _LOGGER.debug(
                "┌─ API REQUEST ──────────────────────────────────────────────"
            )
            _LOGGER.debug("│ Method: %s", method)
            _LOGGER.debug("│ URL: %s", url)
            _LOGGER.debug(
                "│ Headers:\n%s", json.dumps(self._headers_auth, indent=2)
            )
            if "json" in kwargs:
                _LOGGER.debug(
                    "│ Request body (JSON):\n%s",
                    json.dumps(kwargs["json"], indent=2),
                )
            if "data" in kwargs:
                _LOGGER.debug("│ Request body (data): %s", kwargs["data"])
            _LOGGER.debug(
                "└────────────────────────────────────────────────────────────"
            )

        # Basic retry loop for transient errors
        max_attempts = 3
        for attempt in range(1, max_attempts + 1):
            if _LOGGER.isEnabledFor(logging.DEBUG) and attempt > 1:
                _LOGGER.debug(
                    "Retry attempt %d/%d for %s %s",
                    attempt,
                    max_attempts,
                    method,
                    url,
                )
            try:
                async with async_timeout.timeout(self._request_timeout):
                    response = await session.request(
                        method,
                        url,
                        headers=self._headers_auth,
                        **kwargs,
                    )
                    _LOGGER.debug(
                        "SmartCocoon API response status: %s", response.status
                    )

                    # Enhanced debug logging for HA integration
                    if _LOGGER.isEnabledFor(logging.DEBUG):
                        _LOGGER.debug(
                            "┌─ API RESPONSE"
                            " ─────────────────────────────────────────────"
                        )
                        _LOGGER.debug("│ Method: %s | URL: %s", method, url)
                        _LOGGER.debug("│ Status: %s", response.status)
                        _LOGGER.debug(
                            "│ Headers:\n%s",
                            json.dumps(dict(response.headers), indent=2),
                        )

                    response.raise_for_status()
                    data = await response.json(content_type=None)

                    # Debug: Log response body
                    if _LOGGER.isEnabledFor(logging.DEBUG):
                        _LOGGER.debug(
                            "│ Response body:\n%s", json.dumps(data, indent=2)
                        )
                        _LOGGER.debug(
                            "└────────────────────────────────────────────────────────────"  # pylint: disable=line-too-long
                        )
                    break
            except ClientResponseError as err:
                if err.status in (401, 403):
                    _LOGGER.debug(
                        "Auth error (%s), re-authenticating", err.status
                    )
                    # Try to re-authenticate once for protected endpoints
                    if url != API_AUTH_URL and self._bearer_token:
                        # Force auth to refresh token with existing uid/client
                        self._authenticated = False
                        # Raise UnauthorizedError so caller can re-authenticate
                        raise UnauthorizedError(str(err)) from err
                    raise UnauthorizedError(str(err)) from err
                if err.status == 429 and attempt < max_attempts:
                    retry_after = (
                        err.headers.get("Retry-After") if err.headers else None
                    )
                    await asyncio.sleep(
                        self._compute_retry_delay(attempt, retry_after)
                    )
                    continue
                if 500 <= err.status < 600 and attempt < max_attempts:
                    await asyncio.sleep(2 ** (attempt - 1))
                    continue
                raise RequestError(str(err)) from err
            except (ClientConnectionError, asyncio.TimeoutError) as err:
                if attempt < max_attempts:
                    await asyncio.sleep(2 ** (attempt - 1))
                    continue
                raise RequestError(str(err)) from err
            except Exception as err:  # pylint: disable=broad-except
                _LOGGER.exception("API call to SmartCocoon failed")
                raise RequestError(str(err)) from err
            finally:
                if not use_running_session:
                    await session.close()

        # If this request is for authorization, save auth data
        if url == API_AUTH_URL:
            if _LOGGER.isEnabledFor(logging.DEBUG):
                _LOGGER.debug(
                    "┌─ AUTHENTICATION SUCCESS ──────────────────────────────"
                )
                _LOGGER.debug(
                    "│ User ID: %s",
                    (
                        data["data"]["id"]
                        if data and "data" in data
                        else "Unknown"
                    ),
                )
                _LOGGER.debug(
                    "│ Email: %s",
                    (
                        data["data"]["email"]
                        if data and "data" in data
                        else "Unknown"
                    ),
                )
                _LOGGER.debug(
                    "│ Token expires in: %s seconds",
                    response.headers.get("expiry", "Unknown"),
                )
                _LOGGER.debug(
                    "└─────────────────────────────────────────────────────────"  # pylint: disable=line-too-long
                )

            # Check if required headers are present
            if "access-token" not in response.headers:
                _LOGGER.error("Missing access-token in response headers")
                return None
            if "expiry" not in response.headers:
                _LOGGER.error("Missing expiry in response headers")
                return None
            if "client" not in response.headers:
                _LOGGER.error("Missing client in response headers")
                return None

            self._bearer_token = response.headers["access-token"]
            self._bearer_token_expiration = datetime.now() + timedelta(
                seconds=int(response.headers["expiry"]) - 10
            )
            self._api_client = response.headers["client"]

            self._headers_auth["access-token"] = self._bearer_token
            self._headers_auth["client"] = self._api_client

            # Only mark as authenticated if we have valid user data
            if (
                data is not None
                and "data" in data
                and "id" in data["data"]
                and "email" in data["data"]
            ):
                self._headers_auth["uid"] = data["data"]["email"]
                self._user_id = data["data"]["id"]
                self._authenticated = True
            else:
                _LOGGER.error(
                    "Authentication failed: Missing or invalid user data "
                    "in response"
                )
                self._authenticated = False

        if data is not None:
            return cast(dict[str, Any], data)

        _LOGGER.error("Response data is None")
        return None

    async def async_get_fan(self, fan_identifier: int) -> dict | None:
        """Fetch a single fan by internal identifier."""
        return await self.async_request(
            "GET", f"{API_FANS_URL}{fan_identifier}"
        )

    async def async_update_fan(
        self, fan_identifier: int, mode: str, power: int
    ) -> dict | None:
        """Update a fan's mode and power."""
        request_body: dict[str, Any] = {"json": {"mode": mode, "power": power}}
        return await self.async_request(
            "PUT", f"{API_FANS_URL}{fan_identifier}", **request_body
        )



"""SmartCocoon constants."""

from enum import Enum, StrEnum

# API Data
API_URL = "https://app.mysmartcocoon.com/api/"
API_AUTH_URL = API_URL + "auth/sign_in"
API_FANS_URL = API_URL + "fans/"
API_HEADERS = {
    "Content-Type": "application/json",
    "User-Agent": "SmartCocoon/12601 CFNetwork/1568.300.101 Darwin/24.2.0",
}

# Default fan speed is 38% in the SmartCocoon app
# but 33% works better for HomeAssistant
DEFAULT_FAN_POWER_PCT = 33

DEFAULT_TIMEOUT: int = 30


class EntityType(Enum):
    """Class to define entity types"""

    # SmartCocoon entity types

    LOCATIONS = "client_systems"
    THERMOSTATS = "thermostats"
    ROOMS = "rooms"
    FANS = "fans"


class FanMode(StrEnum):
    """Fan mode."""

    ECO = "eco"
    OFF = "always_off"
    ON = "always_on"
    AUTO = "auto"


class FanState(StrEnum):
    """Fan State."""

    FAN_OFF = "false"
    FAN_ON = "true"



#!/usr/bin/env python3
"""Monitor fan connection status to see how long it takes to update."""

import asyncio
import logging
import os
import sys
from datetime import datetime
from pathlib import Path

from dotenv import load_dotenv

from pysmartcocoon import SmartCocoonManager

# Add the parent directory to the path so we can import pysmartcocoon
sys.path.insert(0, str(Path(__file__).parent.parent))

# Load environment variables from .env file
load_dotenv(Path(__file__).parent / ".env")

# Set up logging
logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)


async def monitor_connection_status() -> (
    None
):  # pylint: disable=too-many-branches,too-many-statements
    """Monitor fan connection status over time."""
    print("=" * 80)
    print("🔌 Fan Connection Status Monitor")
    print("=" * 80)
    print()

    # Get credentials from environment variables
    username = os.getenv("USERNAME")
    password = os.getenv("PASSWORD")

    if not username or not password:
        print(
            "❌ Missing credentials! Please set USERNAME and PASSWORD in .env"
        )
        return

    manager = SmartCocoonManager()

    try:
        print("🔐 Authenticating...")
        auth_result = await manager.async_start_services(
            username=username, password=password
        )

        if not auth_result:
            print("❌ Authentication failed!")
            return

        print("✅ Authentication successful!")
        print()

        # Get initial data
        print("📊 Getting initial fan data...")
        await manager.async_update_data()

        if not manager.fans:
            print("❌ No fans found!")
            return

        print(f"✅ Found {len(manager.fans)} fans")
        print()

        # Show initial status
        print("🔍 Initial Connection Status:")
        print("-" * 50)
        for fan_id, fan in manager.fans.items():
            last_conn = fan.last_connection
            if last_conn:
                time_since = datetime.now(last_conn.tzinfo) - last_conn
                time_str = str(time_since).split(".", maxsplit=1)[
                    0
                ]  # Remove microseconds
            else:
                time_str = "Unknown"

            print(f"  Fan {fan_id}:")
            print(f"    Connected: {fan.connected}")
            print(f"    Last Connection: {last_conn}")
            print(f"    Time Since: {time_str}")
            print()

        print("⏰ Starting monitoring loop...")
        print("   Press Ctrl+C to stop")
        print("-" * 50)

        # Monitor loop
        start_time = datetime.now()
        check_count = 0

        while True:
            check_count += 1
            current_time = datetime.now()
            elapsed = current_time - start_time

            print(
                f"\n🔄 Check #{check_count} - "
                f"Elapsed: {str(elapsed).split('.', maxsplit=1)[0]}"
            )

            # Update data
            await manager.async_update_data()

            # Show status for each fan
            for fan_id, fan in manager.fans.items():
                last_conn = fan.last_connection
                if last_conn:
                    time_since = current_time - last_conn.replace(tzinfo=None)
                    time_str = str(time_since).split(".", maxsplit=1)[0]
                else:
                    time_str = "Unknown"

                status_icon = "🟢" if fan.connected else "🔴"
                print(
                    f"  {status_icon} Fan {fan_id}: "
                    f"Connected={fan.connected}, Last seen: {time_str}"
                )

            # Wait 30 seconds before next check
            print("   ⏳ Waiting 30 seconds...")
            await asyncio.sleep(30)

    except KeyboardInterrupt:
        print("\n\n⏹️  Monitoring stopped by user")
    except (ConnectionError, TimeoutError, ValueError) as e:
        print(f"\n❌ Error: {e}")
    finally:
        # Clean up
        if hasattr(
            manager, "_api"
        ) and hasattr(  # pylint: disable=protected-access
            manager._api, "close"  # pylint: disable=protected-access
        ):
            await manager._api.close()  # pylint: disable=protected-access


if __name__ == "__main__":
    asyncio.run(monitor_connection_status())



#!/usr/bin/env python3
"""Test script to test async_update_data() method with debug logging."""

import asyncio
import logging
import os
import sys
from pathlib import Path

import pytest
from dotenv import load_dotenv

from pysmartcocoon import SmartCocoonManager

# Add the parent directory to the path so we can import pysmartcocoon
sys.path.insert(0, str(Path(__file__).parent.parent))

# Load environment variables from .env file
load_dotenv(Path(__file__).parent / ".env")

# Set up logging to show debug output with better formatting
logging.basicConfig(
    level=logging.DEBUG,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)


@pytest.mark.asyncio
@pytest.mark.integration
async def test_update_data() -> (  # pylint: disable=too-many-branches,too-many-statements,duplicate-code
    None
):
    """Test the async_update_data() method which calls all APIs internally."""
    print("=" * 60)
    print("Testing async_update_data() - Comprehensive API Test")
    print("=" * 60)
    print()

    manager = SmartCocoonManager()

    try:
        print("🔐 Step 1: Authenticating...")

        # Get credentials from environment variables
        username = os.getenv("USERNAME")
        password = os.getenv("PASSWORD")

        if not username or not password:
            print("❌ Missing credentials! Please set USERNAME and PASSWORD")
            print("   Example .env file:")
            print("   USERNAME=your_email@example.com")
            print("   PASSWORD=your_password")
            pytest.skip("Missing credentials")

        auth_result = await manager.async_start_services(
            username=username, password=password
        )

        if not auth_result:
            print("❌ Authentication failed!")
            pytest.fail("Authentication failed")

        print("✅ Authentication successful!")
        print()

        print("🔄 Step 2: Testing async_update_data() - calls all APIs...")
        print("   📊 Watch for detailed debug output below:")
        print("   " + "-" * 50)
        print()

        # Test the main update method that calls all APIs
        await manager.async_update_data()

        print()
        print("   " + "-" * 50)
        print("✅ async_update_data() completed!")
        print()

        # Show comprehensive results
        print("📋 Results Summary:")
        print(f"   • Locations: {len(manager.locations)} found")
        print(f"   • Rooms: {len(manager.rooms)} found")
        print(f"   • Thermostats: {len(manager.thermostats)} found")
        print(f"   • Fans: {len(manager.fans)} found")
        print()

        # Show details for each category
        if manager.locations:
            print("   🏠 Locations:")
            for location_id, location in manager.locations.items():
                print(
                    f"     - ID: {location_id}, "
                    f"Identifier: {location.identifier}"
                )
                if hasattr(location, "postal_code"):
                    print(f"       Postal Code: {location.postal_code}")

        if manager.rooms:
            print("   🏠 Rooms:")
            for room_id, room in manager.rooms.items():
                print(f"     - ID: {room_id}, Name: {room.name}")
                print(
                    f"       Temperature: {room.temperature}°C, "
                    f"Target: {room.target_temperature}°C"
                )

        if manager.thermostats:
            print("   🌡️  Thermostats:")
            for thermo_id, thermo in manager.thermostats.items():
                print(f"     - ID: {thermo_id}, Name: {thermo.name}")
                print(
                    f"       Mode: {thermo.hvac_mode}, "
                    f"State: {thermo.hvac_state}"
                )

        if manager.fans:
            print("   💨 Fans:")
            for _fan_id, fan in manager.fans.items():
                # Try to get the name from various possible sources
                fan_name = "Unknown"
                if hasattr(fan, "name"):
                    fan_name = fan.name
                elif (  # pylint: disable=protected-access
                    hasattr(fan, "_raw_data")
                    and isinstance(fan._raw_data, dict)
                    and "name" in fan._raw_data
                ):
                    fan_name = fan._raw_data["name"]
                print(f"     - ID: {fan.fan_id}, Name: {fan_name}")
                print(
                    f"       Room: {fan.room_name}, "
                    f"Connected: {fan.connected}"
                )
                print(f"       Power: {fan.power}W, Speed: {fan.speed_pct}%")

        print()
        print("=" * 60)
        print("Test completed successfully! 🎉")
        print("=" * 60)

    except Exception as e:  # pylint: disable=broad-exception-caught
        print(f"❌ Test failed with error: {e}")
        import traceback  # pylint: disable=import-outside-toplevel

        traceback.print_exc()
        pytest.fail(f"Test failed with error: {e}")
    finally:
        # Clean up
        if hasattr(
            manager, "_api"
        ) and hasattr(  # pylint: disable=protected-access
            manager._api, "close"  # pylint: disable=protected-access
        ):
            await manager._api.close()  # pylint: disable=protected-access


if __name__ == "__main__":
    asyncio.run(test_update_data())



#!/usr/bin/env python3
"""Unit tests that don't require real API calls."""

from pysmartcocoon import SmartCocoonAPI, SmartCocoonManager
from pysmartcocoon.const import API_AUTH_URL, API_URL


def test_constants() -> None:
    """Test that constants are properly defined."""
    assert API_URL == "https://app.mysmartcocoon.com/api/"
    assert API_AUTH_URL == "https://app.mysmartcocoon.com/api/auth/sign_in"


def test_imports() -> None:
    """Test that main classes can be imported."""
    # Test that classes are importable
    assert SmartCocoonAPI is not None
    assert SmartCocoonManager is not None


def test_basic_functionality() -> None:
    """Test basic functionality without API calls."""

    # Create manager instance
    manager = SmartCocoonManager()

    # Test that properties exist
    assert hasattr(manager, "locations")
    assert hasattr(manager, "thermostats")
    assert hasattr(manager, "rooms")
    assert hasattr(manager, "fans")

    # Test that properties return empty dicts initially
    assert not manager.locations
    assert not manager.thermostats
    assert not manager.rooms
    assert not manager.fans



"""Helper functions for fan mode and speed logic."""

from __future__ import annotations

from pysmartcocoon.const import DEFAULT_FAN_POWER_PCT, FanMode


def derive_mode_from_speed(
    current_mode: FanMode, fan_speed_pct: int | None
) -> FanMode:
    """Derive the target mode from a desired speed and the current mode."""
    if fan_speed_pct == 0:
        return FanMode.OFF if current_mode == FanMode.ON else current_mode
    return FanMode.ON if current_mode == FanMode.OFF else current_mode


def resolve_speed(
    current_speed_pct: int,
    fan_mode: FanMode,
    fan_speed_pct: int | None,
) -> int | None:
    """Resolve a valid speed (0-100).

    Returns None if invalid input provided.
    """
    if fan_speed_pct is None:
        if fan_mode == FanMode.OFF:
            return current_speed_pct
        if fan_mode == FanMode.ON and current_speed_pct == 0:
            return DEFAULT_FAN_POWER_PCT
        return current_speed_pct
    if fan_speed_pct < 0 or fan_speed_pct > 100:
        return None
    return fan_speed_pct



#!/usr/bin/env python3
"""Test the extra state attributes functionality."""

import asyncio
import logging
import os
import sys
from pathlib import Path

import pytest
from dotenv import load_dotenv

from pysmartcocoon import SmartCocoonManager

# Add the parent directory to the path so we can import pysmartcocoon
sys.path.insert(0, str(Path(__file__).parent.parent))

# Load environment variables from .env file
load_dotenv(Path(__file__).parent / ".env")

# Set up logging
logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)


@pytest.mark.asyncio
@pytest.mark.integration  # pylint: disable=unknown-option-value
async def test_extra_attributes() -> None:  # pylint: disable=duplicate-code
    """Test the extra state attributes functionality."""
    print("=" * 80)
    print("🔍 Testing Extra State Attributes")
    print("=" * 80)
    print()

    # Get credentials from environment variables
    username = os.getenv("USERNAME")
    password = os.getenv("PASSWORD")

    if not username or not password:
        print(
            "❌ Missing credentials! Please set USERNAME and PASSWORD in .env"
        )
        pytest.skip("Missing credentials")

    manager = SmartCocoonManager()

    try:
        print("🔐 Authenticating...")
        auth_result = await manager.async_start_services(
            username=username, password=password
        )

        if not auth_result:
            print("❌ Authentication failed!")
            pytest.fail("Authentication failed")

        print("✅ Authentication successful!")
        print()

        # Get fan data
        print("📊 Getting fan data...")
        await manager.async_update_data()

        if not manager.fans:
            print("❌ No fans found!")
            pytest.fail("No fans found")

        print(f"✅ Found {len(manager.fans)} fans")
        print()

        # Test extra attributes for each fan
        for fan_id, fan in manager.fans.items():
            print(f"🔍 Fan {fan_id} Extra Attributes:")
            print("-" * 50)

            attributes = fan.get_extra_state_attributes()

            for key, value in attributes.items():
                print(f"  {key}: {value}")

            print()

        print("=" * 80)
        print("✅ Extra attributes test completed!")
        print("=" * 80)

    except Exception as e:  # pylint: disable=broad-exception-caught
        print(f"\n❌ Error: {e}")
        pytest.fail(f"Test failed with error: {e}")
    finally:
        # Clean up
        if hasattr(
            manager, "_api"
        ) and hasattr(  # pylint: disable=protected-access
            manager._api, "close"  # pylint: disable=protected-access
        ):
            await manager._api.close()  # pylint: disable=protected-access


if __name__ == "__main__":
    asyncio.run(test_extra_attributes())



"""Define a SmartCocoon Rocation class."""

# pylint: disable=too-few-public-methods,too-many-instance-attributes

from typing import Any


class Location:  # pylint: disable=too-many-instance-attributes
    """Define the location."""

    def __init__(self, data: dict[str, Any]) -> None:
        """Initialize."""
        self._identifier: int = data["id"]
        self._postal_code: str = data["location"]["postal_code"]

    @property
    def identifier(self) -> int:
        """Return location id"""
        return self._identifier

    @property
    def postal_code(self) -> str:
        """Return postal code"""
        return self._postal_code



"""TypedDicts for SmartCocoon API payloads."""

from typing import NotRequired, TypedDict


class FanPayload(TypedDict):
    """Fan payload fields returned by API."""

    id: int
    fan_id: str
    mode: str
    fan_on: bool
    firmware_version: str
    is_room_estimating: bool
    connected: bool
    last_connection: NotRequired[str]
    power: int
    predicted_room_temperature: float
    room_id: int
    thermostat_vendor: int
    mqtt_username: str
    mqtt_password: str


class FansResponse(TypedDict):
    """Response shape for listing fans."""

    fans: list[FanPayload]
